;
; chkout.asm - CHKOUT kernal extension handler
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;    This file is part of IDEDOS the IDE64 disk operating system
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;**************************************************
; Chkout (X=logical file number)
;
; Used registers: A X (Y preserved)
; Status: C
; C=1:
; A=3 file not open
; A=5 device not present
; A=7 (not output file)
; Status response:
; ST=$00 if no error
; ST=$03 timeout
; ST=$80 if device not present
; Secondary address 128..255 then communication on serial bus is a bit different...
;**************************************************
newchkout	.proc
		cpx #0
		beq chkout_notop
		jsr search_logical_X_st
		beq chkout_cont1
chkout_notop	jmp exit_file_not_open

chkout_cont1	jsr readfiletables
		jsr search_buffer_A
		beq chkout_cont2	;Ok found buffer for file

		#ldatax open_files_currentp_exp
		lda open_files_info_exp,x
		ldx #MISCOUT
		sta buffer_info_exp,x
		cmp #INFO_KERNAL
		bne chkout_cont4

		lda ActualDev
		sta buffer_drive_exp,x
		cmp #3			;screen?
		bne *+4
		ldx #STDOUT
		.if (CHRIN_CHROUT_RAM=1) & (SUPERCPU=0)
		jsr chkout_doch
		.else
		#stx_l_ buffer_output_exp
		.fi

		cmp #4
		bcs cw			;serial device?
		#psr $f25d
		lda ActualDev
		jmp exit2basic

cw		ldx #3
cy		cmp current_floppy8_exp,x
		bne xg
		txa
		ora #8
		ldx #0
xg		dex
		bpl cy

		tax
                
                .if (JIFFYDOS_ACCEL=1) & (SUPERCPU=0) & (JIFFYDOS_ACCEL_CHRINOUT=0)
                lda JiffyDos_internal_exp
                cmp #1
                .fi
                
		jsr OP_cfg16up
		txa
                .if JIFFYDOS_ACCEL_CHRINOUT=1
		jsr ieclisten
                .else
                
                .if (JIFFYDOS_ACCEL=1) & (SUPERCPU=0)
                php
                bcs nx
                jsr $ed0c
                jmp nx2
nx              jsr ieclisten
nx2
                .else
                jsr $ed0c
                .fi

                .fi
		lda ActualSecAddr
		bpl sfc
                .if (JIFFYDOS_ACCEL=1) & (SUPERCPU=0) & (JIFFYDOS_ACCEL_CHRINOUT=0)
                plp
                .fi
		jsr $edbe		;clear atn
		jmp xf
sfc
		.if JIFFYDOS_ACCEL_CHRINOUT=1
		jsr iecsecond
                .else
                
                .if (JIFFYDOS_ACCEL=1) & (SUPERCPU=0)
		plp
                bcs nx3
                jsr $edb9
                jmp nx4
nx3             jsr iecsecond
nx4
                .else
                jsr $edb9
                .fi
                
                .fi
xf		jsr OP_cfg16down
		txa
		bit Status
		bmi exit_device_not_present
		sta ActualOutput
		clc
		jmp exit2basic		;to $f25d


chkout_cont4	lda ActualDev
		sta buffer_drive_exp,x

chkout_cont3	.if (CHRIN_CHROUT_RAM=1) & (SUPERCPU=0)
		jsr chkout_doch
		.else
		#stx_l_a buffer_output_exp
		.fi

		.if INPUT_OUTPUT_COMPATIBILITY=1
		lda ActualDev
		sta ActualOutput	;Lamer
		.fi

		clc
		jmp exit2basic

chkout_cont2	txa
		cmp buffer_output_exp	;execute command, if leaving channel
		beq chkout_cont3
		pha
		#ldatax buffer_output_exp
		sta buffer_currentp_exp	;OP_do_command_channel
		lda buffer_info_exp,x
		cmp #INFO_STATUS
		bne chkout_nstatus
		#phy_a
		jsr OP_save_zeropage
		jsr OP_XIT_do_command_channel
		jsr OP_restore_zeropage
		#ply_a
chkout_nstatus	#plx_a
		jmp chkout_cont3
		.pend

		.if (CHRIN_CHROUT_RAM=1) & (SUPERCPU=0)
chkout_doch	.proc
		php
		pha
		txa
		sta buffer_output_exp
		clc
		adc #<buffer_info_exp
		sta newchrout_ram_at_exp
		txa
		clc
		adc #<buffer_address_low_exp
		sta newchrout_ram_at2_exp
		txa
		clc
		adc #<buffer_address_hi_exp
		sta newchrout_ram_at3_exp
		pla
		plp
		rts
		.pend
		.fi
