;
; chrin.asm - CHRIN, GETIN kernal extension handler
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;    This file is part of IDEDOS the IDE64 disk operating system
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;**************************************************
; Chrin & getin
;
; Used registers: A (X Y preserved)
; Status: C (N Z for Accu)
; C=1:
; A=6 (not input file)
; Status response:
; ST=$00 if no error
; ST=$40 if last byte of file, or end of line on channel #15, or end of record.
; ST=$42 if no more bytes (open error, output file, device...)
; ST=$52 if no more bytes, because of readerror (open error, output file, device...)
; ST=$80 if device not present
;
;Note:
;Directory summ at end will be incorrect if multiple read at same time...
;And directory is half done anyway.
;**************************************************
;Buffer:
;
;current file position: (buffer_address_low_exp + buffer_address_hi_exp << 8 + 1) & 0x7ff | buffer_filepos2_exp <<8 ...
;if buffer_address_end_hi_exp == 0 then
;       buffer contains no data, have to read first
;       buffer_address_low/hi_exp must point before the buffer
;	buffer_address_end_low_exp must be 0
;
;if buffer_address_end_hi_exp is after the end of buffer then
;	the whole buffer is needed and there are more sectors to come
;
;if buffer_address_hi_exp & 0x80 then
;       file ends in this memory page
;       buffer_address_end_low_exp selects last byte of buffer
;
;buffer_address_low/hi_exp is only before buffer if it contains no data! so buffer_address_end_low/hi_exp=0
;buffer_address_low/hi_exp points to the previous byte read from buffer if any
;
; file position pointer does not advance if eof...
;--------------------------------------
; GETIN
;--------------------------------------
		.if (CHRIN_CHROUT_RAM=1) & (SUPERCPU=0)
fake		= newchrin_ram
fake2		= newgetin
		.fi

newgetin	.proc
		.if SUPERCPU=1
		phx
		.else
		stx help_x_exp
		.fi
		#ldatax buffer_input_exp
		lda buffer_info_exp,x
		bpl newchrin.getin
back		cmp #INFO_KERNAL
		bne newchrin.chrin_not_data_sector
                .if SUPERCPU=1
                pla
                .fi
		#psr $f140
		.if SUPERCPU=1
                pha
                .fi
		cpx #STDIN
		bne nonstandard
		lda buffer_output_exp
		eor #STDOUT
		bne nonstandard
		.if SUPERCPU=1
                plx
                .fi
		#psr switch_norm
		#psr gotobank3		
                lda #$00
bbb             .if SUPERCPU=0
		ldx help_x_exp
                .fi
		ora #$00                ;Now we switch to bank3 to prevent
                jmp exit2basic          ;raster bugs when calling $ffe4
                                        ;instantly to scan keyboard...
nonstandard	lda buffer_drive_exp,x	
		.if SUPERCPU=1
                plx
                .fi
		cmp #4
                bcc bbb
                .if SUPERCPU=0
		ldx help_x_exp
                .fi
		pla
                pla
	        gcs newchrin.getinen
;--------------------------------------
; CHRIN
;--------------------------------------
newchrin	.proc
		.if SUPERCPU=1
		phx
		.else
		stx help_x_exp
		.fi
		#ldatax buffer_input_exp
		lda buffer_info_exp,x
		bmi chrin_not_data_sector

;BLK_RW_DIRTY, BLK_READ, BLK_RW

getin		.if SUPERCPU=1
		lda $a6
		pha
		lda #1
		sta $a6			;SCPU Page 1
		clc
hard_chrin	lda buffer_address_low_exp,x
		inc a
		sta buffer_address_low_exp,x
		.else
hard_chrin	asr #$00
		isb buffer_address_low_exp,x
		eor #$ff
		.fi
		beq chrin_next_bank
		sta $a4
		lda buffer_address_hi_exp,x
		bmi chrin_data_sector_last
chrin_bank_nen2	sta $a5
chrin_bank_nend

		.if SUPERCPU=1		;Get byte
chrin_lastb
		lda [$a4]
		plx			;Restore $a6
		stx $a6
chrin_v		plx
		.else
chrin_lastb	ldx #0

		lda ($a4,x)
chrin_v		ldx help_x_exp
		.fi
		ora #$00		;Set NZ flags
		jmp exit2basic
;---------------------------------
chrin_next_bank	sta $a4
		.if SUPERCPU=1
		lda buffer_address_hi_exp,x
		inc a
		sta buffer_address_hi_exp,x
		.else
		isb buffer_address_hi_exp,x
		eor #$ff
		.fi
		cmp buffer_address_end_hi_exp,x
		bcc chrin_bank_nen2
		ora #$80
		sta buffer_address_hi_exp,x;Mark next bank

chrin_data_sector_last
		and #$7f
		sta $a5
		lda buffer_address_low_exp,x
		cmp buffer_address_end_low_exp,x
		bcc chrin_bank_nend
		gcs chrin_next
;---------------------------------------

;BLK_WRITE_DIRTY, BLK_RW_EOF_DIRTY, BLK_R_EOF, BLK_RW_EOF, BLK_RW_EOF_FULL, BLK_WRITE_FULL, BLK_WRITE, STATUS, KERNAL, IGNORE

chrin_not_data_sector
		pha
		lda buffer_drive_exp,x
		sta current_drive_exp
		tax
		pla
		cmp #INFO_KERNAL
		bne nchrin_stand

		cpx #4
		bcc vew			;serial device?

		.if SUPERCPU=1
                plx
		.else
		ldx help_x_exp
                .fi

getinen		lda Status
		beq sok
		lda #$0d
		clc
		gcc serr

sok		
		.if JIFFYDOS_ACCEL_CHRINOUT=1
		jsr OP_cfg16up
		jsr iecin
		jsr OP_cfg16down
                lda $a4
                jmp exit2basic
                .else

                .if (JIFFYDOS_ACCEL=1) & (SUPERCPU=0)
                lda JiffyDos_internal_exp
                beq nx
                jsr OP_cfg16up
		jsr iecin
		jsr OP_cfg16down
                lda $a4
                jmp exit2basic
                .fi

nx              #psr $ee13
                jmp exit2basic
                .fi

vew		.if SUPERCPU=1
		txa
		plx
		#psr $f159
                ora #$00
serr		jmp exit2basic
		.else
                #psr $f159
                txa
		ldx help_x_exp
                ora #$00
serr            jmp exit2basic
		.fi
;---------------------------------------

;BLK_WRITE_DIRTY, BLK_RW_EOF_DIRTY, BLK_R_EOF, BLK_RW_EOF, BLK_RW_EOF_FULL, BLK_WRITE_FULL, BLK_WRITE, STATUS, IGNORE

nchrin_stand	cmp #INFO_STATUS
		bne nchrin_status

		#stx_l_a current_drive_exp
		lda status_pos_exp,x
		bpl ok_st		;No pending request

		.if SUPERCPU=1
		phy
		and #$7f
		.else
		tya
		pha
		lda status_pos_exp,x
		and #$7f
		.fi
		tay
		jsr OP_save_zeropage
		#OP_scpu01up
		tya
		jsr LWOP_disk_status
		#lda0
		jsr append_status_dec
		#lda0
		jsr append_status_dec
		#lda0
		jsr append_status_dec
		#lda0
		jsr append_status_dec
                txa
                and #63
                tay
		#OP_scpu01down
		jsr OP_restore_zeropage
		#ldatax current_drive_exp
                tya
                sta status_size_exp,x
		#ply_a

ok_st		lda status_pos_exp,x
		cmp status_size_exp,x
                bcs gon
		txa
		lsr
		ror			;*64
		ror
		ora status_pos_exp,x
		.if SUPERCPU=1
		pha
		and #$3f
		inc a
		sta status_pos_exp,x
		plx
		.else
		inc status_pos_exp,x
		tax
		.fi
		lda drive_status_exp,x
		gcc chrin_v		;end of message? (c=0)

gon     	#ldatax current_drive_exp
		lda #$80		;00, OK,000,000,000
		sta status_pos_exp,x
		lsr			;lda #$40
		sta Status
		lda #$0d
					;clc
		gcc chrin_v
;---------------------------------------

;BLK_WRITE_DIRTY, BLK_RW_EOF_DIRTY, BLK_R_EOF, BLK_RW_EOF, BLK_RW_EOF_FULL, BLK_WRITE_FULL, BLK_WRITE, IGNORE

nchrin_status	lda #$42		;All the rest...
		sta Status
nchrin_err_end	clc
		.if SUPERCPU=1
		plx
		.else
		ldx help_x_exp
		.fi
		#lda0			;$c7
		jmp exit2basic
;---------------------------------------
chrin_next	#stx_l_a buffer_currentp_exp
		lda buffer_drive_exp,x
		sta current_drive_exp
		#phy_a
                #OP_scpu01up
		jsr OP_save_zeropage

		jsr XIT_do_next
		bcc *+6
		lda #$52
		sta Status

		jsr OP_restore_zeropage
                #OP_scpu01down
		#ply_a

		lda #%10111111
		bit Status
		bvs chrin_lastb
		bne nchrin_err_end
		#ldatax buffer_input_exp
		gcc hard_chrin
;---------------------------------------
;BLK_RW_DIRTY, BLK_READ, BLK_RW
;
XIT_do_next	#xit_init

		#ldatax buffer_currentp_exp
		lda buffer_type_exp,x
		cmp #BUFFERMODE_FILE
		beq XIT_do_filenext
		cmp #BUFFERMODE_DIR
		beq XIT_do_dirnext
		cmp #BUFFERMODE_DIRECT
		beq XIT_do_direct
                cmp #BUFFERMODE_PCLINK
                beq XIT_do_pclink
		brk
;---------------------------------------
XIT_do_filenext	lda buffer_info_exp,x
		pha
		lda #INFO_IGNORE	;be safe
		sta buffer_info_exp,x

		lda buffer_address_end_hi_exp,x
		beq dc
		lda $a5			;buffer_address_hi_exp,x
		cmp buffer_address_max_hi_exp,x
		beq dc2			;eof mark?

		lda #$40
		sta Status

		pla
		cmp #INFO_BLK_READ
		bne dc3
		lda #INFO_BLK_R_EOF
		pha
		gcs mkx			;No flush, just exit

dc3		cmp #INFO_BLK_RW
		beq v2
		jsr XIT_flush_buf	;-=> error
v2		lda #INFO_BLK_RW_EOF
		pha
mkx		#ldatax buffer_currentp_exp
		jmp chrin_chrout_do_file_end.more

dc2		pla
		cmp #INFO_BLK_RW_DIRTY
		bne v1
		jsr XIT_flush_buf	;-=> error
		lda #INFO_BLK_RW
v1		pha

dp		#ldatax buffer_currentp_exp
		lda buffer_address_max_hi_exp,x
		sec
		sbc buffer_start,x
		clc
		adc buffer_filepos2_exp,x
		sta buffer_filepos2_exp,x
		bcc dc
		#inc_l_x_a buffer_filepos3_exp
		bne dc
		#inc_l_x_a buffer_filepos4_exp;!!!!! (4Gb)
dc
		jsr XIT_load_tree	;-=> error

		jsr XIT_read_in_buffer	;-=> error

		jmp chrin_chrout_do_file_end
;---------------------------------------
XIT_do_dirnext	ldx #write_byte_r2end-write_byte_r2-1
htdo2		lda write_byte_r2,x
		sta start_zeropage,x
		dex
		bpl htdo2

		#ldatax buffer_currentp_exp
		lda buffer_info_exp,x
		pha
		lda #INFO_IGNORE	;be safe
		sta buffer_info_exp,x

		jsr OP_init_buffer_pointers

		lda buffer_start,x
		sta start_zeropage+6
		jsr OP_load_next_sector_now

		lda #>b_dir_misc_exp
		jsr buffer_subpart_cp

;                ldy #<b_dir_misc_exp
		ldx #direntry_backup_end_exp-direntry_backup_start_exp-1
nvw		#lda_zp_y current_path
		sta direntry_backup_start_exp,x
		iny
		dex
		bpl nvw

		lda direntry_first_exp
		bpl ndirlastbyte

		pla			;INFO_IGNORE
do_eof		lda #$40
		sta Status
		clc
		rts			;-=> last byte of file

ndirlastbyte	lda #>b_dir_searchstring_exp
		jsr buffer_subpart_cp	;restore searchstring

		jsr XIT_dirlp2		;-=> error

		lda direntry_first_exp
		bpl hq
		lda start_zeropage+5
		bne *+4
		dec start_zeropage+6
		dec start_zeropage+5
hq
		#ldatax buffer_currentp_exp
		lda start_zeropage+5
		sta buffer_address_end_low_exp,x
		lda start_zeropage+6
		sta buffer_address_end_hi_exp,x

		lda #>b_dir_misc_exp
		jsr buffer_subpart_cp

;		ldy #<b_dir_misc_exp
		ldx #direntry_backup_end_exp-direntry_backup_start_exp-1
nvw2		lda direntry_backup_start_exp,x
		#sta_zp_y current_path
		iny
		dex
		bpl nvw2

		#ldatax buffer_currentp_exp
		jsr OP_save_next_sector_now
		pla			;be safe
		sta buffer_info_exp,x
		clc			;-=> ok
		rts
;---------------------------------------
XIT_do_direct	jsr OP_init_buffer_pointers
		jmp do_eof

XIT_do_pclink	#ldatax buffer_currentp_exp
		lda buffer_info_exp,x
		pha
		lda #INFO_IGNORE	;be safe
		sta buffer_info_exp,x

                lda buffer_address_end_hi_exp,x
		beq dcv
		lda $a5			;buffer_address_hi_exp,x
		cmp buffer_address_max_hi_exp,x
		beq dcv			;eof mark?
                
                lda #$40
                sta Status
                pla
                lda #INFO_BLK_R_EOF
                pha
                #jmps chrin_chrout_do_file_end.more

dcv             jsr pclink_chrin
                pha
                jsr OP_init_buffer_pointers
                pla
                beq chrin_chrout_do_file_end.more
                tya
                bne *+4
                dec pom2
                dey
		lda pom2
                #jmps chrin_chrout_do_file_end.fer
		.pend
;---------------------------------------
; Common end of both chrin and chrout
;
chrin_chrout_do_file_end .proc
		jsr OP_init_buffer_pointers

		lda buffer_filesize1_exp,x
		sec
		sbc #1
		tay
		lda buffer_filesize2_exp,x
		sbc buffer_filepos2_exp,x
		sta pom2
		lda buffer_filesize3_exp,x
		sbc buffer_filepos3_exp,x
		bne more
		lda buffer_filesize4_exp,x
		sbc buffer_filepos4_exp,x
		bcc more
		bne more
		lda buffer_address_max_hi_exp,x
		sec
		sbc buffer_start,x
		clc
		sbc pom2
		bcc more

		lda pom2
		ora buffer_start,x
		gcs fer

more		ldy #0
		lda buffer_address_max_hi_exp,x
fer		sta buffer_address_end_hi_exp,x
		tya
		sta buffer_address_end_low_exp,x

		pla			;be safe
		sta buffer_info_exp,x
		clc
		rts			;-=> ok
		.pend

write_byte_r2	.proc
		.logical start_zeropage
		.if SUPERCPU=1
		txa			;1
		tyx			;2
		tay			;3
lp		pla			;4
		sta $1ff00,x		;8
		dex			;9
		bpl lp			;11
		tyx			;12
		rts			;13
		.else
		jmp lp			;3
lp		pla			;4
		sta $ff00,y		;7
		dey			;8
		bpl lp			;10
		rts			;11
		.fi
		.here
		.pend
write_byte_r2end

		.if (CHRIN_CHROUT_RAM=1) & (SUPERCPU=0)

chrin_data_sector_last__
		stx help_x_exp
		ldx buffer_input_exp
		lda newchrin_ram.tmpp+1
		sta $a4
		lda buffer_address_hi_exp,x
		jmp newchrin.chrin_data_sector_last

chrin_next_bank__
		stx help_x_exp
		ldx buffer_input_exp
		jmp newchrin.chrin_next_bank

chrin_not_data_sector__
		stx help_x_exp
		ldx buffer_input_exp
		jmp newchrin.chrin_not_data_sector

newchrin_ram_from
		.logical newchrin_ram_exp
newchrin_ram	.proc
		lda buffer_info_exp+STDIN;3
		bmi chrin_not_data_sector_;5

getin		asr #0			;7

		isb buffer_address_low_exp	;10
		eor #$ff		;12
		beq chrin_next_bank_	;14
		sta tmpp+1		;17
		lda buffer_address_hi_exp;20
		bmi chrin_data_sector_last_;22
		sta tmpp+2		;25

tmpp		lda $ffff		;28
		jmp exit2basic		;31

chrin_not_data_sector_
		jmp chrin_not_data_sector__;34
chrin_next_bank_
		jmp chrin_next_bank__	;37
chrin_data_sector_last_
		jmp chrin_data_sector_last__;40

newgetin_ram	.proc
		lda buffer_info_exp+STDIN;43
		bpl newchrin_ram.getin	;45
		jmp newgetin_back__	;48

newchrout_ram	.proc
		sta bce+1		;51
		lda buffer_info_exp+STDOUT;54
		asr #$01       		;56
		bcs chrout_not_data_sector_;58

		isb buffer_address_low_exp	;61
		eor #$ff		;63
		beq chrout_nxt_bank_	;65
		sta tmppp+1		;68
		lda buffer_address_hi_exp;71
		bmi chrout_data_sector_last_;73
		sta tmppp+2		;76

bce		lda #0			;78
tmppp		sta $ffff		;81
		jmp exit2basic		;84

chrout_not_data_sector_
		jmp chrout_not_data_sector__;87
chrout_nxt_bank_
		jmp chrout_nxt_bank__	;90
chrout_data_sector_last_
		jmp chrout_data_sector_last__;93
		.pend

		.here

newgetin_back__	stx help_x_exp
		ldx buffer_input_exp
		jmp newgetin.back

chrout_not_data_sector__
		stx help_x_exp
		ldx buffer_output_exp
		lda newchrout_ram.bce+1
		pha
		jmp newchrout.chrout_not_data_sector

chrout_nxt_bank__
		stx help_x_exp
		ldx buffer_output_exp
		lda newchrout_ram.bce+1
		pha
		lda #0
		jmp newchrout.chrout_nxt_bank

chrout_data_sector_last__
		stx help_x_exp
		ldx buffer_output_exp
		lda newchrout_ram.bce+1
		pha
		lda newchrout_ram.tmppp+1
		sta $a4
		lda buffer_address_hi_exp,x
		jmp newchrout.chrout_data_sector_last
		.fi
