;
; open.asm - OPEN kernal extension handler
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;    This file is part of IDEDOS the IDE64 disk operating system
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;**************************************************
; Open
;
; Used registers: A X Y
; Status: C
; C=1:
; A=1 too many files
; A=2 file open
; A=5 device not present, and filename is given
; A=6 not input file
; A=7 (not output file)
; Status response:
; ST=$00 allways, except:
; ST=$80 if device not present, and filename is given
;
; Notes:
; If logical number=0 then err(not input file)
; If logical number exists then err(file open)
; If open_files>9 then err(too many files)
; Secondary address 128..255 then filename is ignored
;**************************************************

;--------------------------------------
; OPEN
;--------------------------------------
newopen		.proc
                .if DEBUGIO=1
                lda debug_io_exp
                beq kd
                #lda0
                sta debug_io_exp
                #psr $ffc3
                php
                pha
                #phx_a
                #phy_a
                lda #3
                jmp monitor
kd		lda #1
		sta debug_io_exp
                .fi

		ldx ActualLogical
		beq exit_not_input_file

		jsr search_logical_X_st	;ST=$00
		beq exit_file_open

		#lda0
		jsr search_logical_A	;search empty entry
		bne exit_too_many_files

		lda ActualSecAddr
		ora #$60
		sta ActualSecAddr
		sta open_files_secondary_exp,x
		lda ActualDev
		sta open_files_device_exp,x
                beq kernal_open.vd
		#devicehandler hard_open,lcd_open,pclink_open2,kernal_open
		.pend

kernal_open	.proc
		tya
		beq vd
		cmp #31
		bcs open_ignore
		cmp #3
		beq vd
		bcc al
		lda ActualSecAddr
		bmi vd
		lda ActualFNLength
		beq vd
		#stz_a Status

		#psr $f382
		#psr $f3e3

		#ldatax open_files_currentp_exp
		lda #INFO_KERNAL
		sta open_files_info_exp,x
		lda ActualLogical
		sta open_files_logical_exp,x;Now it's officially open
		tya
		jmp exit2basic

al		#psr $f384		;Go to CBM kernal...
vd		lda #INFO_KERNAL	;Must be a standard CBM device...
		.byte SKIP2

open_ignore	.proc
		lda #INFO_IGNORE

open_ok_mode	.proc
		#ldx_l_ open_files_currentp_exp
		sta open_files_info_exp,x
		lda ActualLogical
openback	.proc
		sta open_files_logical_exp,x;Now it's officially open
		lda ActualDev
		clc
		jmp exit2basic
		.pend

open_buffer_ok_mode .proc
		#ldx_l_ buffer_currentp_exp
		sta buffer_info_exp,x
		lda ActualLogical
		sta buffer_logical_exp,x;Now it's officially allocated
		#ldx_l_ open_files_currentp_exp
		#jmps openback
		.pend

hard_open	.proc
		#lda0
		jsr search_buffer_A
		bne open_ignore		;All buffers taken???

		jsr OP_copy_filename_to_path

		#ldatax buffer_currentp_exp
		lda current_drive_exp
		sta buffer_drive_exp,x

		lda ActualSecAddr
		bmi open_ignore
		and #$0f
		eor #$0f
		bne open_std_file

		sta buffer_address_low_exp,x
                lda #BUFFERMODE_ERROR
                sta buffer_type_exp,x

		lda ActualFNLength
		beq hard_open_nerror

		#psr hard_open.back
		#psr switch_16k
		jmp hard_open_nerror

hard_open.back	#sta_cfg_open
		pha
                #OP_scpu01up
		jsr OP_save_zeropage
		jsr cfe
		jsr OP_restore_zeropage
                #OP_scpu01down
		clc
		pla
		jmp exit2basic

hard_open_nerror
		lda #INFO_STATUS
		gne open_buffer_ok_mode

open_std_file	lda ActualFNLength
		beq open_ignore

		#OP_scpu01up
		jsr OP_save_zeropage
		jsr XIT_open
		pha
		jsr OP_restore_zeropage
                #OP_scpu01down
		pla
		bcs open_ignore
		gcc open_buffer_ok_mode


cfe		#ldatax buffer_currentp_exp
		jmp OP_XIT_execute_15command_open

XIT_open	#xit_init

		jsr OP_get_current_drive_type
		and #$c0
		cmp #$80		;cdrom?
		php

		jsr OP_init_buffer_pointers
		#lda0
		sta buffer_address_end_hi_exp,x
		sta buffer_address_end_low_exp,x;Auto read first sector
		lda #>512
		plp
		bne *+4
		lda #>2048
		clc
		adc buffer_start,x
		sta buffer_address_max_hi_exp,x;Size of buffer

		lda path_exp
		cmp #"$"
		beq open_dir
		cmp #"#"
		beq open_direct

		jsr check_drive		;check drive is alive

		jsr XIT_init_drive

		lda #RESOLV_OPENFIRST_MODE
		jsr OP_XIT_resolv_path_part;-=> error
		php

		ldy #4
ds		lda path_searchtype_exp
		cmp invl1,y
		bne vr
		lda path_searchtype_exp+1
		cmp invl2,y
		bne vr
		lda path_searchtype_exp+2
		cmp invl3,y
		bne vr			;sorry invalid filetype -> prg
gon		sec
		gcs LWOP_reporterror_64_xit;file type mismatch
vr		dey
		bpl ds

		lda resolv_anything_exp
		sec
		beq LWOP_reporterror_62_xit

		#ldatax buffer_currentp_exp
		#lda0
		sta buffer_filepos2_exp,x
		sta buffer_filepos3_exp,x
		sta buffer_filepos4_exp,x
		lda #BUFFERMODE_FILE
		sta buffer_type_exp,x

		lda resolv_open_mode_exp
		cmp #"R"
		beq open_file_read
		cmp #"W"
		sec
		bne LWOP_reporterror_64_xit;not implemented yet
		plp
		bcs open_file_write
		sec
                lda path_rewrite_exp
		beq LWOP_reporterror_63_xit;file exists

                #phx_a			;it's not the most optimal way for "@:", but it works
                jsr rewrite_file	;remove file
                ldx #15
np              lda direntry_f_name_exp,x
                sta path_exp,x
                dex
                bpl np
                #stx_l_a resolv_empty_entry_found_exp
                ldx #3
np2             lda direntry_f_filetype_exp-1,x
                sta path_searchtype_exp-1,x
                dex
                bne np2
                #stx_l_a path_exp+16
                #plx_a
                #jmps open_file_write

invl1		.text "DRCDL"
invl2		.text "EEBIN"
invl3		.text "LLMRK"
;---------------------------------------
open_file_write	#stx_l_a buffer_currentp_exp

                #lda0
		sta buffer_filesize1_exp,x
		sta buffer_filesize2_exp,x
		sta buffer_filesize3_exp,x
		sta buffer_filesize4_exp,x
                
                lda setup_atribut2_exp
                asl

		lda buffer_start,x
		ora #>b_cfs_file_bitmap_exp
                ldx #0
                bcs *+4
                ldx #"/"
		jsr XIT_prealloc_dir_sector

		jsr find_free_sector	;allocate tree sector

		#cnext next_sector , created_new_tree_sector_exp;remember tree sector position

		jsr OP_save_next_sector_treep;save tree sector

		lda #>b_cfs_file_misc_exp
		jsr buffer_subpart_cp

		ldy #<(b_cfs_file_treeroot_exp+3);backup treeroot
		ldx #4
uiy		lda next_sector-1,x
		#sta_zp_y current_path
		dey
		dex
		bne uiy

		ldy #<b_cfs_file_treepos_exp
		txa
		#sta_zp_y current_path	;set the current treepos to 0
		iny
		#sta_zp_y current_path

		lda #>b_cfs_file_tree_exp
		jsr buffer_subpart_cp
		tya
vs		#sta_zp_y current_path	;zero out tree sector
		iny
		bne vs
		inc current_path+1
vs2		#sta_zp_y current_path
		iny
		bne vs2

		#stz_a szam+3		;filelen=0
		#stz_aa szam+2
		#stz_aa szam+1
		#stz_aa szam

		lda #FTYPE_NORMAL | FATTR_EXEC | FATTR_READ | FATTR_WRITE | FATTR_DELETE
		jsr XIT_finalize_direntry	;-=> error

		lda #>b_cfs_file_misc_exp
		jsr buffer_subpart_cp
		ldy #<b_cfs_file_dirsect_exp
		ldx #3
lc		lda next_sector,x
		#sta_zp_y current_path
		iny
		dex
		bpl lc
		ldy #<b_cfs_file_dir_offs_exp
                lda resolv_entry_offs_exp
		#sta_zp_y current_path

		jsr save_allocator_state

		clc
		lda #INFO_BLK_WRITE
		rts			;-=> ok
;---------------------------------------
open_file_read	pla			;pop

		lda direntry_f_size_exp
		sta buffer_filesize1_exp,x
		lda direntry_f_size_exp+1
		sta buffer_filesize2_exp,x
		lda direntry_f_size_exp+2
		sta buffer_filesize3_exp,x
		lda direntry_f_size_exp+3
		sta buffer_filesize4_exp,x

                jsr OP_get_current_drive_type
                and #FS_MASK
                cmp #FS_CFS
                beq open_file_read_cfs
                sec
                gcs LWOP_reporterror_62_xit;only cfs is supported now

open_file_read_cfs
		lda #>b_cfs_file_misc_exp
		jsr buffer_subpart_cp

		ldy #<(b_cfs_file_treeroot_exp+3);backup treeroot
		ldx #3
uiy2		lda next_sector,x
		#sta_zp_y current_path
		dey
		dex
		bpl uiy2

		ldy #<b_cfs_file_treepos_exp
		lda #$ff
		#sta_zp_y current_path	;set the current treepos to 65535!!!!!!!!!
		iny
		#sta_zp_y current_path

		clc
		lda #INFO_BLK_READ
inv     	rts			;-=> ok
;---------------------------------------
open_dir	lda ActualSecAddr
		and #$0f
		beq ms			;0 is normal directory
		lsr
                beq inv                 ;1 is invalid
		lda #1			;2-14 is raw
ms		pha
		jsr check_drive		;check drive is alive

		jsr XIT_init_drive	;-=> error

		pla
		ora #2*2
		jsr XIT_dir_first	;-=> error

		#ldatax buffer_currentp_exp
		lda #BUFFERMODE_DIR
		sta buffer_type_exp,x
		jsr OP_save_next_sector_now

		lda #>b_dir_searchstring_exp;backup searchstring
		jsr buffer_subpart_cp
		ldx #0
hthon		lda path_exp,x
		#sta_zp_y current_path
		beq mfw
		iny
		inx
		bne hthon
mfw
		lda #>b_dir_type_format_exp;backup type
		jsr buffer_subpart_cp
		ldx #0
hthon2		lda path_searchtype_exp,x
		#sta_zp_y current_path
		beq mfw2
		iny
		inx
		bne hthon2
mfw2
		lda #>b_dir_misc_exp
		jsr buffer_subpart_cp

;		ldy #<b_dir_misc_exp
		ldx #direntry_backup_end_exp-direntry_backup_start_exp-1
nvw		lda direntry_backup_start_exp,x
		#sta_zp_y current_path
		iny
		dex
		bpl nvw

		clc
		lda #INFO_BLK_READ
		rts			;-=> ok
;---------------------------------------
open_direct
		lda ActualSecAddr
		and #$0f
		cmp #2
		bcs ma			;channels 2-14 are valid only
		sec
		rts
ma
		jsr OP_init_buffer_pointers
		lda buffer_address_max_hi_exp,x
                sec
                sbc #1
		sta buffer_address_end_hi_exp,x
		lda #255
		sta buffer_address_end_low_exp,x
		lda #BUFFERMODE_DIRECT
		sta buffer_type_exp,x

		clc
		lda #INFO_BLK_RW_DIRTY
		rts			;-=> ok
		.pend


OP_init_buffer_pointers .proc
		#ldatax buffer_currentp_exp
		lda buffer_start,x
		sec
		sbc #1
		sta buffer_address_hi_exp,x
		lda #255
		sta buffer_address_low_exp,x
		rts
		.pend
;----------------------------------
search_logical_X_st .proc
		#stz_a Status
		txa
;----------------------------
search_logical_A .proc
		ldx #MAX_OPEN_FILES
lp		dex
		bmi exit
		cmp open_files_logical_exp,x
		bne lp
		#stx_l open_files_currentp_exp
exit		rts
		.pend
;-------------------------
search_buffer_A	.proc
		ldx #MAX_BUFFERS
lp		dex
		bmi exit
		cmp buffer_logical_exp,x
		bne lp
		#stx_l buffer_currentp_exp
exit		rts
		.pend
;-------------------------
readfiletables	.proc
		pha
		lda open_files_logical_exp,x
		sta ActualLogical
		lda open_files_device_exp,x
		sta ActualDev
		lda open_files_secondary_exp,x
		sta ActualSecAddr
		pla
		rts
		.pend
;-------------------------
pclink_open2	.proc
		#lda0
		jsr search_buffer_A
		bne open_ignore2       	;All buffers taken???

		jsr OP_init_buffer_pointers
		#lda0
		sta buffer_address_end_hi_exp,x
		sta buffer_address_end_low_exp,x;Auto read first sector
		lda #>2048
		clc
		adc buffer_start,x
		sta buffer_address_max_hi_exp,x;Size of buffer
                lda #BUFFERMODE_PCLINK
                sta buffer_type_exp,x

		lda ActualSecAddr
                sta buffer_pclink_secaddr,x
		bmi open_ignore2

		jsr OP_save_zeropage
		jsr OP_copy_filename_to_path
		jsr pclink_open
                pha
                jsr OP_restore_zeropage
                pla

		bcs open_ignore2
		gcc open_buffer_ok_mode
                
open_ignore2	jmp open_ignore
		.pend

lcd_open	.proc
		.if LCD_DSPL=1
		lda ActualSecAddr
		and #$0f
		eor #$01		;Channel 1 write, others read
		bne open_ignore
		lda #INFO_CHAR_WRITE
		gne open_ok_mode
		.if INFO_CHAR_WRITE=0
		.error INFO_CHAR_WRITE=0?
		.fi
		.fi
